/*
 * Copyright (C) 2009 by Eric Lambert <Eric.Lambert@sun.com>
 * Use and distribution licensed under the
 * GNU Lesser General Public License (LGPL) version 2.1.
 * See the COPYING file in the parent directory for full text.
 */
package org.gearman.client;

import org.gearman.Constants;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

import org.gearman.common.GearmanPacket;
/**
 * This interface exposes the API which users of the Gearman Java Libaray will
 * use to submit jobs for execution by the Gearman System and to track the
 * progess of those jobs.
 *
 * <p>
 * Users of a GearmanClient will first instantiate a <tt>GearmanClient</tt>
 * instance. After the instance has been created, the user will need to
 * register the set of Gearman Job Servers which are to be availble to client
 * for job submission.
 *
 * <p>
 * Once the <tt>GearmanClient</tt> has been configured with the appropriate
 * set of Gearman Job Servers, the user may begin using the <tt>GearmanClient</tt>
 * to submit jobs for execution. This is accomplished by passing a
 * {@link GearmanJob} to either the
 * {@link GearmanClient#submit(java.util.concurrent.Callable) } or the
 * {@link GearmanClient#invokeAll(java.util.Collection) } methods. Both of
 * these methods return a {@link Future} object which can be used to track
 * the state of that job.
 *
 * <p>
 * A <tt>GearmanClient</tt> can be shut down, which will cause it
 * to stop accepting new tasks.  After being shut down, the client
 * will eventually terminate, at which point no tasks are actively
 * executing, no tasks are awaiting execution, and no new tasks can be
 * submitted
 *
 *  <h3>Usage Example</h3>
 *
 * Here is an example of how to use a <tt>GearmanClient</tt> to submit and
 * monitor the execution of a {@link GearmanJob} (Note, this example does not
 * show the creation of the GearmanClient instance nor the creation of the
 * GearmanJob as these details are implementation specific):
 *
 * <pre>
 *     org.gearman.GearmanClient client;
 *     org.gearman.GearmanJob job;
 *
 *     // Here you would instantiate the client and job
 *
 *     client.addJobServer("localhost",4730);
 *
 *     Future<GearmanJobResult> jobFuture = client.submit(job);
 *
 *     GearmanJobResult result = jobFuture.get();
 *     ...
 *
 * </pre>
 *
 */
public interface GearmanClient extends ExecutorService {

    /**
     * Register a new Gearman Job Server with the client. Assumes the Job Server
     * is listening on the default port {@link Constants#GEARMAN_DEFAULT_TCP_PORT}.
     *
     * @param host The name or IP Address of the host running the Gearman Job
     *             Server.
     * @throws IllegalArgumentException If an invalid host has been specified.
     * @throws IOException
     * @throws IllegalStateException If the client has already been stopped.
     */
    void addJobServer(String host) throws IllegalArgumentException,
            IOException, IllegalStateException;

    /**
     * Register a new Gearman Job Server with the client.
     * @param host The name or IP Address of the host running the Gearman Job
     *             Server.
     * @param port The port on which the Gearman Job Server is listening.
     * @throws IllegalArgumentException If an invalid host has been specified.
     * @throws IOException
     * @throws IllegalStateException If the client has already been stopped.
     */
    void addJobServer(String host, int port)
            throws IllegalArgumentException, IOException, IllegalStateException;

    /**
     * Determine if the client has registered a Gearman Job Server running on 
     * the specified host and listening on the default port {@link Constants#GEARMAN_DEFAULT_TCP_PORT}.
     * 
     * @param host The name or IP Address of the host running the Gearman Job
     *             Server.
     * @return true if the Job Server has been registered with the client, else
     *         false.
     * @throws IllegalArgumentException If an invalid host has been specified.
     * @throws IllegalStateException If the client has already been stopped.
     */
    boolean hasJobServer(String host) throws IllegalArgumentException,
            IllegalStateException;

    /**
     * Determine if the client has registered a Gearman Job Server running on
     * the specified host and listening on the specified port.
     *
     * @param host The name or IP Address of the host running the Gearman Job
     *             Server.
     * @param port The port on which the Gearman Job Server is listening.
     * @return true if the Job Server has been registered with the client, else
     *         false.
     * @throws IllegalArgumentException If an invalid host has been specified.
     * @throws IllegalStateException If the client has already been stopped.
     */
    boolean hasJobServer(String host, int port)
            throws IllegalArgumentException, IllegalStateException;

    /**
     * Retrieve the list of Gearman Job Servers that have been registered
     * with this client.
     *
     * @return the list of registered Gearman Job Servers.
     *
     * @throws IllegalStateException If the client has already been stopped.
     */
    List<InetSocketAddress> getSetOfJobServers()
            throws IllegalStateException;

    /**
     * Unregisters a Gearman Job Server from use by this client.
     * @param host The name or IP Address of the host running the Gearman Job
     *             Server.
     * @throws IllegalArgumentException If an invalid host has been specified.
     * @throws IllegalStateException If the client has already been stopped or
     *         the host is currently processing one or more requests.
     */
    void removeJobServer(String host) throws IllegalArgumentException,
            IllegalStateException;

     /**
     * Unregisters a Gearman Job Server from use by this client.
     * @param host The name or IP Address of the host running the Gearman Job
     *             Server.
     * @param port The port on which the Gearman Job Server is listening.
     * @throws IllegalArgumentException If an invalid host has been specified.
     * @throws IllegalStateException If the client has already been stopped or
     *         the host is currently processing one or more requests.
     */
    void removeJobServer(String host, int port)
            throws IllegalArgumentException, IllegalStateException;

    /**
     * As {@link GearmanJob} are submitted and executed, they will receive
     * notifications from the Gearman Job Server as certain events occur (such
     * as JOB_COMPLETE or WORK_DATA). Users may want to process these events,
     * for example if jobs are sending back intermediate data, a user may wish
     * to receive that data and act upon it. This method returns to the user
     * all events which have occured for any job submitted by the client.
     *
     * @return A collection of {@link  GearmanJobEvent} that have been received.
     * @throws IllegalStateException If the client has already been stopped.
     */
    Collection<GearmanPacket> selectUpdatedJobEvents()
            throws IllegalStateException;

    /**
     * Sends a WORK_STATUS request to the Gearman Job Server running the
     * {@link GearmanJob} and then returns {@link GearmanJobStatus} representing
     * the current status of the job.
     *
     * @param job The job for which status request is being made.
     * @return The status of the specified job.
     * @throws IOException
     * @throws IllegalStateException If the client has already been stopped.
     */
    GearmanJobStatus getJobStatus(GearmanJob job) throws IOException,
            IllegalStateException;
}
