/*
 * Copyright (C) 2009 by Eric Lambert <Eric.Lambert@sun.com>
 * Use and distribution licensed under the
 * GNU Lesser General Public License (LGPL) version 2.1.
 * See the COPYING file in the parent directory for full text.
 */
package org.gearman.client;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;import org.gearman.Constants;
import org.gearman.GearmanException;
import org.gearman.common.GearmanPacket;
import org.gearman.common.GearmanJobServerSession;
import org.gearman.common.GearmanServerRequest;
import org.gearman.common.GearmanServerResponseHandler;
import org.gearman.common.Packet;
import org.gearman.common.PacketMagic;
import org.gearman.common.PacketType;
import org.gearman.util.ByteUtils;

//TODO change the server selection to use open connection


/* TODO
 * -documentation (specifically the unsupported methods)
 * -handle dropped sessions
 * -have updateJobStatus handle any type of job (currently only handles jobimpl)
 */

/*
 * ISSUES/RFEs
 * -several of the invoke/submit/execute methods throw OpNotSupported exception
 *  because they rely task cancelation, which we dont support
 * -shutdownNow will always return an empty set, there is a few problems here
 *  1) a gearmanJob is a callable, not a runnable
 *  2) we track requests, not jobs inside the session object
 * -selectUpdate should provide the ability to on select on certain events
 */
public class GearmanClientImpl implements GearmanClient, GearmanJobEventListener {

    private static enum state {

        RUNNING, SHUTTINGDOWN, TERMINATED
    }
    private static final String DESCRIPION_PREFIX = "GearmanClient";
    private final String DESCRIPTION;
    private HashMap<SelectionKey, GearmanJobServerSession> sessionsMap = null;
    private ArrayList<GearmanJobServerSession> sessionList = null;
    private Selector ioAvailable = null;
    private ArrayList<GearmanPacket> updatedJobs = null;
    private static final Logger LOG = Logger.getLogger(
            Constants.GEARMAN_CLIENT_LOGGER_NAME);
    private state runState = state.RUNNING;

    public static Logger getLogger() {
        return LOG;
    }

    /**
     * Create a new GearmanClient instance. Instances are not thread-safe and
     * should not be shared across threads.
     */
    public GearmanClientImpl() {
        sessionsMap = new HashMap<SelectionKey, GearmanJobServerSession>();
        sessionList = new ArrayList<GearmanJobServerSession>();
        updatedJobs = new ArrayList<GearmanPacket>();
        DESCRIPTION = new String(DESCRIPION_PREFIX + ":" +
                Thread.currentThread().getId());
    }

    public void addJobServer(String host) throws IllegalArgumentException,
            IOException, IllegalStateException {
        addJobServer(host, Constants.GEARMAN_DEFAULT_TCP_PORT);
    }

    public void addJobServer(String host, int port)
            throws IllegalArgumentException, IOException,
            IllegalStateException {

        if (!runState.equals(state.RUNNING)) {
            throw new RejectedExecutionException("Client has been shutdown");
        }

        InetSocketAddress sa = new InetSocketAddress(host, port);

        //this is a sub-optimal way to look for dups, but addJobServer
        //ops should be infrequent enough that this should be a big penalty
        Iterator<GearmanJobServerSession> iter = sessionList.iterator();
        while (iter.hasNext()) {
            if (iter.next().getServer().equals(sa)) {
                return;
            }
        }

        GearmanJobServerSession session = new GearmanJobServerSession(sa);
        sessionList.add(session);

        if (ioAvailable == null) {
            ioAvailable = Selector.open();
        }
        LOG.log(Level.FINE, "Added server " + sa + " to client " + this);

    }

    public boolean hasJobServer(String host) throws IllegalArgumentException,
            IllegalStateException {
        return hasJobServer(host, Constants.GEARMAN_DEFAULT_TCP_PORT);
    }

    public boolean hasJobServer(String host, int port)
            throws IllegalArgumentException, IllegalStateException {
        if (!runState.equals(state.RUNNING)) {
            throw new IllegalStateException("Client is not active");
        }
        return getSetOfJobServers().contains(new InetSocketAddress(host, port));
    }

    public List<InetSocketAddress> getSetOfJobServers()
            throws IllegalStateException {
        if (!runState.equals(state.RUNNING)) {
            throw new IllegalStateException("Client is not active");
        }

        ArrayList<InetSocketAddress> retSet = new ArrayList<InetSocketAddress>();
        Iterator<GearmanJobServerSession> iter = sessionList.iterator();
        while (iter.hasNext()) {
            retSet.add(iter.next().getServer());
        }
        return retSet;
    }

    public void removeJobServer(String host) throws IllegalArgumentException,
            IllegalStateException {
        removeJobServer(host, Constants.GEARMAN_DEFAULT_TCP_PORT);
    }

    public void removeJobServer(String host, int port)
            throws IllegalArgumentException, IllegalStateException {
        if (!runState.equals(state.RUNNING)) {
            throw new IllegalStateException("JobServers can not be removed " +
                    "once shutdown has been commenced.");
        }

        //TODO, make this better
        InetSocketAddress sa = new InetSocketAddress(host, port);
        Iterator<GearmanJobServerSession> iter = sessionList.iterator();
        GearmanJobServerSession session = null;
        boolean foundit = false;
        while (iter.hasNext() && !foundit) {
            session = iter.next();
            if (session.getServer().equals(sa)) {
                foundit = true;
            }
        }

        if (!foundit) {
            throw new IllegalArgumentException("JobServer " + sa + " has not" +
                    " been registered with this client.");
        }

        if (session.isProcessingRequests()) {
            throw new IllegalStateException("Can not remove GearmanJobServer " +
                    sa + ". JobServer is still processing jobs from this" +
                    " client");
        }

        if (session.isInitialized()) {
            SelectionKey k = session.getSelectionKey();
            if (k != null) {
                sessionsMap.remove(k);
                k.cancel();
            }
            session.closeSession();
        }
        sessionList.remove(session);
        LOG.log(Level.FINE, "Removed job server " + sa + " from client " + this);
    }

    public <T> Future<T> submit(Callable<T> task) {

        if (task == null) {
            throw new NullPointerException("Null task was submitted to " +
                    "gearman client");
        }

        if (!runState.equals(state.RUNNING)) {
            throw new RejectedExecutionException("Client has been shutdown");
        }

        if (!(task instanceof GearmanServerResponseHandler)) {
            throw new RejectedExecutionException("Task must implement the " +
                    GearmanServerResponseHandler.class + " interface to" +
                    " submitted to this client");
        }

        GearmanJobImpl  job = (GearmanJobImpl) task;
        GearmanServerResponseHandler handler = (GearmanServerResponseHandler) job;


        if (job.isDone()) {
            throw new RejectedExecutionException("Task can not be resubmitted ");
        }

        GearmanJobServerSession session = null;
        try {
            session = getSessionForTask();
        } catch (IOException ioe) {
            throw new RejectedExecutionException(ioe);
        }
        job.setJobServerSession(session);
        job.registerEventListener(this);
        GearmanPacket submitRequest = getPacketFromJob(job);
        GearmanServerRequest submittedJob = new GearmanServerRequest(handler, submitRequest);
        session.submitRequest(submittedJob);
        LOG.log(Level.FINE, "Client " + this + " has submitted job " + job +
                " to session " + session + ". Job has been added to the " +
                "active job queue");
        try {
            driveRequestTillState(session, submittedJob,
                    GearmanServerRequest.RequestState.PROCESSINGRESPONSES, 10);
        } catch (IOException ioe) {
            LOG.log(Level.WARNING, "Client " + this + " encounted an IOException" +
                    " while drivingIO");
        }

        return (Future) job;
    }

    public <T> Future<T> submit(Runnable task, T result) {
        throw new UnsupportedOperationException("Client does not support " +
                "execution of non-GearmanJob objects");
    }

    public Future<?> submit(Runnable task) {
        throw new UnsupportedOperationException("Client does not support " +
                "execution of non-GearmanJob objects");
    }

    public void execute(Runnable command) {
        throw new UnsupportedOperationException("Client does not support " +
                "execution of non-GearmanJob objects");
    }

    public <T> List<Future<T>> invokeAll(Collection<Callable<T>> tasks)
            throws InterruptedException {
        ArrayList<Future<T>> futures = new ArrayList();
        for (Callable curTask : tasks) {
            futures.add(this.submit(curTask));
        }
        for (Future<T> results : futures) {
            try {
                results.get();
            } catch (ExecutionException ee) {
                //TODO
            }
        }
        return futures;

    }

    public <T> List<Future<T>> invokeAll(Collection<Callable<T>> tasks,
            long timeout, TimeUnit unit) throws InterruptedException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public <T> T invokeAny(Collection<Callable<T>> tasks)
            throws InterruptedException, ExecutionException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public <T> T invokeAny(Collection<Callable<T>> tasks, long timeout,
            TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public GearmanJobStatus getJobStatus(GearmanJob job) throws IOException,
            GearmanException, IllegalStateException {
        // we need to know what session to ask that information can not be
        // gleened from the job object.
        if (!(job instanceof GearmanJobImpl)) {
            throw new IllegalArgumentException("job must be of type " +
                    GearmanJobImpl.class);
        }
        GearmanJobImpl jobImpl = (GearmanJobImpl) job;
        return updateJobStatus(jobImpl.getHandle(), jobImpl.getSession());
    }

    public byte[] echo(byte[] data) throws IOException, GearmanException {
        if (!runState.equals(state.RUNNING)) {
            throw new IllegalStateException("Client is not active");
        }
        GearmanPacket echoRequest = new Packet(PacketMagic.REQ, PacketType.ECHO_REQ,
                data);
        GearmanEchoResponseHandler handler = new GearmanEchoResponseHandler();
        GearmanServerRequest t = new GearmanServerRequest(handler, echoRequest);
        GearmanJobServerSession session = getSessionForTask();
        session.submitRequest(t);
        LOG.log(Level.FINE, "Client " + this + " has submitted echo request " +
                "(payload = " + ByteUtils.toHex(data) +
                " to session " + session);
        if (!driveRequestTillState(session, t, GearmanServerRequest.RequestState.FINISHED, 10)) {
            throw new GearmanException("Failed to execute echo request " + t +
                    " to session " + session);
        }
        LOG.log(Level.FINE, "Client " + this + " has completed echo request " +
                "to session " + session);
        return handler.getResults();
    }

    public Collection<GearmanPacket> selectUpdatedJobEvents() throws
            GearmanException, IllegalStateException {
        if (!runState.equals(state.RUNNING)) {
            throw new IllegalStateException("Client is not active");
        }
        Collection<GearmanPacket> retSet = new HashSet<GearmanPacket>();
        try {
            driveClientIO();
        } catch (IOException ioe) {
            LOG.log(Level.WARNING, "Encountered IOException while driving " +
                    "client IO" + ioe);
            //TODO improve ioexception handling
        }
        if (!updatedJobs.isEmpty()) {
            retSet = (Collection<GearmanPacket>) updatedJobs.clone();
            updatedJobs.clear();
        }
        return retSet;
    }

    public int getNumberofActiveJobs() throws IllegalStateException {
        if (runState.equals(state.TERMINATED)) {
            throw new IllegalStateException("Client is not active");
        }
        int numJobs = 0;
        for (GearmanJobServerSession session : sessionList) {
            if (session.isInitialized()) {
                numJobs += session.getNumberOfActiveRequests();
            }
        }
        return numJobs;
    }

        public void handleGearmanJobEvent(GearmanPacket event) throws IllegalArgumentException {
            if (event == null) {
                throw new IllegalArgumentException("event can not be null");
            }

            PacketType type = event.getPacketType();
            switch (type) {
                case JOB_CREATED:
                    updatedJobs.add(event);
                    break;
                case WORK_STATUS:
                    updatedJobs.add(event);
                    break;
                case WORK_DATA:
                    updatedJobs.add(event);
                    break;
                case WORK_WARNING:
                    updatedJobs.add(event);
                    break;
                case WORK_COMPLETE:
                    updatedJobs.add(event);
                    break;
                case WORK_FAIL:
                    updatedJobs.add(event);
                    break;
                case WORK_EXCEPTION:
                    updatedJobs.add(event);
                    break;
                default:
                    LOG.log(Level.WARNING, "Client " + this + " received an " +
                            "un-recognized job event that it can not handle: " +
                            event);
            }

    }

    public void shutdown() {
        if (!runState.equals(state.RUNNING)) {
            return;
        }
        runState = state.SHUTTINGDOWN;
        LOG.log(Level.FINE, "Commencing controlled shutdown of client: " + this);
        try {
            awaitTermination(-1, TimeUnit.SECONDS);
        } catch (InterruptedException ie) {
            LOG.log(Level.FINE, "Client shutdown interrupted while waiting" +
                    " for jobs to terminate.");
        }
        shutdownNow();
        LOG.log(Level.FINE, "Completed ontrolled shutdown of client: " + this);
    }

    public List<Runnable> shutdownNow() {

        runState = state.SHUTTINGDOWN;
        LOG.log(Level.FINE, "Commencing immediate shutdown of client: " + this);
        Iterator<GearmanJobServerSession> sessions = sessionList.iterator();
        while (sessions.hasNext()) {
            GearmanJobServerSession curSession = sessions.next();
            if (!curSession.isInitialized()) {
                continue;
            }
            try {
                curSession.closeSession();
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Failed to closes session " + curSession +
                        " while performing immediate shutdown of client " +
                        this + ". Encountered the following exception " + e);
            }
            sessions.remove();
        }
        sessionsMap.clear();
        sessionsMap = null;
        sessionList.clear();
        sessionList = null;
        updatedJobs.clear();
        updatedJobs = null;
        runState = state.TERMINATED;
        LOG.log(Level.FINE, "Completed shutdown of client: " + this);
        return new ArrayList();
    }

    public boolean isShutdown() {
        return !runState.equals(state.RUNNING);
    }

    public boolean isTerminated() {
        return runState.equals(state.TERMINATED);
    }

    public boolean awaitTermination(long timeout, TimeUnit unit)
            throws InterruptedException {
        TimeUnit sessionUnit = TimeUnit.MILLISECONDS;
        long timeLeft = -1;
        long timeOutInMills = timeout < 0 ? -1 : TimeUnit.MILLISECONDS.convert(timeout, unit) + System.currentTimeMillis();

        if (getNumberofActiveJobs() == 0) {
            return true;
        }

        for (GearmanJobServerSession curSession : sessionList) {
            if ((!curSession.isInitialized())
                    || curSession.getNumberOfActiveRequests() == 0) {
                continue;
            }
            //negative timeout means block till completion
            if (timeout >= 0) {
                timeLeft = timeOutInMills - System.currentTimeMillis();
                if (timeLeft <= 0) {
                    LOG.log(Level.WARNING, "awaitTermination exceeded timeout.");
                    break;
                }
            }
            try {
                curSession.waitForRequestsToComplete(timeLeft, sessionUnit);
            } catch (TimeoutException te) {
                LOG.log(Level.FINE, "timed out waiting for all tasks to complete");
                break;
            }
        }
        return getNumberofActiveJobs() == 0;
    }

    public String toString() {
        return DESCRIPTION;
    }

    private void driveClientIO() throws IOException, GearmanException {
        ioAvailable.selectNow();
        Set<SelectionKey> keys = ioAvailable.selectedKeys();
        LOG.log(Level.FINEST, "Driving IO for client " + this + ". " +
                keys.size() + " session(s) currently available for IO");
        Iterator<SelectionKey> iter = keys.iterator();
        while (iter.hasNext()) {
            SelectionKey key = iter.next();
            GearmanJobServerSession s = sessionsMap.get(key);
            s.driveSessionIO();
        }
    }

    private GearmanJobStatus updateJobStatus(byte[] jobhandle,
            GearmanJobServerSession session) throws IOException,
            IllegalStateException, GearmanException {
        if (!runState.equals(state.RUNNING)) {
            throw new IllegalStateException("Client is not active");
        }

        if (jobhandle == null || jobhandle.length == 0) {
            throw new IllegalStateException("Invalid job handle. Handle must" +
                    " not be null nor empty");
        }
        GearmanPacket statusRequest = new Packet(PacketMagic.REQ,
                PacketType.GET_STATUS, jobhandle);
        GearmanServerResponseHandler handler = (GearmanServerResponseHandler)
                new GearmanJobStatusImpl();
        GearmanServerRequest t = new GearmanServerRequest(
                handler, statusRequest);
        session.submitRequest(t);
        if (!driveRequestTillState(session, t, GearmanServerRequest.RequestState.FINISHED, 10)) {
            throw new GearmanException("Failed to execute jobstatus request " +
                    t + " to session " + session);
        }
        return (GearmanJobStatus)handler;
    }

    private GearmanJobServerSession getSessionForTask() throws IOException {
        if (sessionList.isEmpty()) {
            throw new IOException("No servers registered with client");
        }

        int s = (int) Math.round((Math.random() * (sessionList.size() - 1)));
        GearmanJobServerSession session = sessionList.get(s);
        if (!session.isInitialized()) {
            session.initSession(ioAvailable,SelectionKey.OP_READ | SelectionKey.OP_WRITE );
            SelectionKey key = session.getSelectionKey();
            sessionsMap.put(key, session);
        }
        return session;
    }

    private boolean driveRequestTillState(GearmanJobServerSession session,
            GearmanServerRequest r, GearmanServerRequest.RequestState stat,
            int attempts)
            throws IOException, GearmanException {
        while (r.getState().compareTo(stat) < 0 && attempts != 0) {
            session.driveSessionIO();
            if (attempts > 0) {
                attempts--;
            }
            try {
                Thread.sleep(100);
            } catch (InterruptedException ie) {
                break;
            }
        }
        return r.getState().compareTo(stat) >= 0;
    }

    private GearmanPacket getPacketFromJob(GearmanJob job) {
        int destPos = 0;
        PacketMagic magic = PacketMagic.REQ;
        PacketType type = null;
        byte[] packetdata = null;
        byte[] fnname = ByteUtils.toAsciiBytes(job.getFunctionName());
        byte[] uid = job.getID();
        byte[] data = job.getData();
        if (job.getPriority().equals(GearmanJob.JobPriority.HIGH)) {
            type = job.isBackgroundJob() ? PacketType.SUBMIT_JOB_HIGH_BG : PacketType.SUBMIT_JOB_HIGH;
        }
        if (job.getPriority().equals(GearmanJob.JobPriority.LOW)) {
            type = job.isBackgroundJob() ? PacketType.SUBMIT_JOB_LOW_BG : PacketType.SUBMIT_JOB_LOW;
        }
        if (job.getPriority().equals(GearmanJob.JobPriority.NORMAL)) {
            type = job.isBackgroundJob() ? PacketType.SUBMIT_JOB_BG : PacketType.SUBMIT_JOB;
        }
        packetdata = new byte[fnname.length + uid.length + data.length + 2];
        System.arraycopy(fnname, 0, packetdata, destPos, fnname.length);
        destPos += fnname.length;
        packetdata[destPos++] = ByteUtils.NULL;
        System.arraycopy(uid, 0, packetdata, destPos, uid.length);
        destPos += uid.length;
        packetdata[destPos++] = ByteUtils.NULL;
        System.arraycopy(data, 0, packetdata, destPos, data.length);
        return new Packet(magic, type, packetdata);
    }
}
