/*
 * Copyright (C) 2009 by Eric Lambert <Eric.Lambert@sun.com>
 * Use and distribution licensed under the
 * GNU Lesser General Public License (LGPL) version 2.1.
 * See the COPYING file in the parent directory for full text.
 */
package org.gearman.common;



import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Queue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.gearman.GearmanException;
import org.gearman.client.GearmanClientImpl;
import org.gearman.client.GearmanJobImpl;
// TODO HANDLE CONNECTION DROPS/IO FAILURES
// TODO javadoc
public class GearmanJobServerSession {

    static final String DESCRIPTION_PREFIX = "GearmanJobServerSession";
    private final String DESCRIPTION;
    private InetSocketAddress remote = null;

    // all request currently being handled by this session
    private Queue<GearmanServerRequest> sessionRequests = null;
    // all request that have  been attached to this session but have yet to be
    // passed on the server
    private Queue<GearmanServerRequest> newRequests = null;
    private Queue<GearmanPacket> responsesReceived = null;
    private static final Logger LOG = GearmanClientImpl.getLogger();
    private GearmanNIOJobServerConnection connection = null;
    private SelectionKey sessionSelectionKey = null;

    public GearmanJobServerSession(InetSocketAddress sa) {
        if (sa == null) {
            throw new IllegalArgumentException("Session can be created with " +
                    "a null remote address");
        }
        remote = sa;
        connection = new GearmanNIOJobServerConnection(remote);
        DESCRIPTION = new String(DESCRIPTION_PREFIX + ":" +
                Thread.currentThread().getId() + ":" + sa.toString());
    }

    public String toString() {
        return DESCRIPTION;
    }

    public void initSession(Selector sel, int mask) throws IllegalStateException,IOException {
        if (isInitialized()) {
            throw new IllegalStateException("A session can not be " +
                    "initialized twice");
        }
        connection.open();
        responsesReceived = new LinkedList<GearmanPacket>();
        newRequests = new LinkedList<GearmanServerRequest>();
        sessionRequests = new LinkedList<GearmanServerRequest>();
        LOG.log(Level.FINE, "Session " + this + " has been initialized.");
        sessionSelectionKey = connection.registerSelector(sel, mask);
    }

    public InetSocketAddress getServer() {
        InetSocketAddress ret = null;
        if (remote != null) {
            ret = new InetSocketAddress(remote.getHostName(), remote.getPort());
        }
        return ret;
    }

    public SelectionKey getSelectionKey () {
        if (!isInitialized()) {
            throw new IllegalStateException("Session " + this +
                    " has not been initialized");
        }
        return sessionSelectionKey;
    }

    public int getNumberOfActiveRequests() throws IllegalStateException {
        if (!isInitialized()) {
            throw new IllegalStateException("Session " + this +
                    " has not been initialized");
        }
        return sessionRequests.size();
    }

    public boolean isInitialized() {
        return (connection != null && connection.isInitialized());
    }

    public void waitForRequestsToComplete() throws IllegalStateException, InterruptedException {
        if (!isInitialized()) {
            throw new IllegalStateException("Session " + this +
                    " has not been initialized");
        }
        try {
            waitForRequestsToComplete(-1, TimeUnit.SECONDS);
        } catch (TimeoutException ignore) {
        } //should never happen
    }

    public void waitForRequestsToComplete(long timeout, TimeUnit unit)
            throws InterruptedException, TimeoutException, IllegalStateException {
        if (!isInitialized()) {
            throw new IllegalStateException("Session " + this +
                    " has not been initialized");
        }
        long timeOutInMills = timeout < 0 ? -1 : TimeUnit.MILLISECONDS.convert(timeout, unit) + System.currentTimeMillis();
        while (sessionRequests.size() != 0 &&
                !(timeOutInMills < 0 ? false : System.currentTimeMillis() > timeOutInMills)) {
            Thread.sleep(100);
        }
        if (sessionRequests.size() > 0) {
            throw new TimeoutException("Session " + this + " timed out " +
                    "waiting for all requests to complete");
        }
    }

    public void closeSession() throws IllegalStateException {
        if (!isInitialized()) {
            throw new IllegalStateException("Can not close a session that " +
                    "has not been initialized");
        }
        LOG.log(Level.FINE, "Session " + this + " is being closed.");
        sessionSelectionKey.cancel();
        connection.close();
        responsesReceived.clear();
        responsesReceived = null;
        newRequests.clear();
        newRequests = null;
        sessionRequests.clear();
        sessionRequests = null;
        connection = null;
        LOG.log(Level.FINE, "Session " + this + " has successfully closed.");
    }

    public void submitRequest(GearmanServerRequest request)
            throws IllegalStateException {
        if (!isInitialized()) {
            throw new IllegalStateException("Session hasnt been initialized." +
                    " Request may not be submitted at this time");
        }
        if (request == null) {
            throw new IllegalStateException("A null request can not be" +
                    " submitted to a server");
        }

        if (!request.getState().equals(GearmanServerRequest.RequestState.NEW)) {
            throw new IllegalStateException("Invalid request state: " +
                    request.getState());
        }

        //add request to new requests
        newRequests.add(request);
        LOG.log(Level.FINE, "Session " + this + " is now handling " +
                "the request " + request);
    }

    public void sendRequestToServer(GearmanServerRequest request) throws IllegalStateException {
        if (!isInitialized()) {
            throw new IllegalStateException("Session " + this +
                    " has not been initialized");
        }
        newRequests.add(request);
    }

    public boolean isProcessingRequests() {
        if (sessionRequests == null && newRequests == null) {
            return false;
        }
        return ((!sessionRequests.isEmpty()) || (!newRequests.isEmpty()));
    }

    public void driveSessionIO() throws IOException, GearmanException, IllegalStateException {
        if (!isInitialized()) {
            throw new IllegalStateException("you can not driveSessionIO on an" +
                    " un-initialized session");
        }

        while (connection.canWrite() && !newRequests.isEmpty()) {
                GearmanServerRequest request = newRequests.remove();
                GearmanPacket p = request.getRequestPacket();
                connection.write(p);
                request.handleGearmanJobEvent(p);
                if (!request.getState().equals(GearmanServerRequest.RequestState.FINISHED)) {
                    sessionRequests.add(request);
                }
                LOG.log(Level.FINE, "Session " + this + " successfully " +
                        "submitted request " + request + " to its job server");
        }

        while (connection.canRead()) {
                GearmanPacket p = connection.read();
                if (p == null) {
                    continue;
                }
                GearmanServerRequest curReq = matchResponseWithRequest(p);
                if (curReq != null) {
                    curReq.handleGearmanJobEvent(p);
                    if (curReq.getState().equals(GearmanServerRequest.RequestState.FINISHED) ||
                            curReq.getState().equals(GearmanServerRequest.RequestState.EXCEPTION)) {
                        sessionRequests.remove(curReq);
                    }
                } else {
                    if (p.getPacketType().equals(PacketType.NOOP)) {
                        LOG.log(Level.WARNING,"Received a NOOP and cant find a presleep to associate with it");
                    }
                }
        }
    }

    private GearmanServerRequest matchResponseWithRequest(GearmanPacket response) {
        GearmanServerRequest request = null;
        if (response.getPacketType().equals(PacketType.ERROR)) {
            LOG.log(Level.SEVERE, "Receieved an error packet from Server: " + response + ". Shutting down session");
            return request;
            //TODO Implement shutdown
        }
        for (GearmanServerRequest req : sessionRequests) {
            if (isAnonymousResposnse(response.getPacketType()) &&
                    req.getState().equals(GearmanServerRequest.RequestState.AWAITINGRESPONSE)) {
                request = req;
                break;
            } else {
                byte[] handle = extractJobHandle(response.getData());
                if (req.getHandler() instanceof GearmanJobImpl) {
                    GearmanJobImpl job = (GearmanJobImpl) req.getHandler();
                    if (!Arrays.equals(handle, job.getHandle())) {
                        continue;
                    }
                    request = req;
                    break;
                }
            }
        }
        if (request == null) {
            LOG.log(Level.WARNING, "Unable to match repsonse " + response +
                    " with a request");
        } else {
            LOG.log(Level.FINE, "Repsonse " + response + " has been matched " +
                    "to request " + request);
        }
        return request;
    }

    private static byte[] extractJobHandle(byte[] data) {
        byte[] handle = new byte[0];
        int i = 0;
        for (i = 0; i < data.length; i++) {
            if (data[i] == (byte) 0) {
                break;
            }
        }
        if (i != data.length) {
            handle = new byte[i];
            System.arraycopy(data, 0, handle, 0, i);
        }
        return handle;
    }

    public boolean isAnonymousResposnse(PacketType type) {
        if (type.equals(PacketType.JOB_CREATED) ||
                type.equals(PacketType.ECHO_RES) ||
                type.equals(PacketType.OPTION_RES) ||
                type.equals(PacketType.STATUS_RES) ||
                type.equals(PacketType.NOOP) ||
                type.equals(PacketType.NO_JOB) ||
                type.equals(PacketType.JOB_ASSIGN) ||
                type.equals(PacketType.JOB_ASSIGN_UNIQ)) {
            return true;
        } else {
            return false;
        }
    }


}
