    /*
 * Copyright (C) 2009 by Eric Lambert <Eric.Lambert@sun.com>
 * Use and distribution licensed under the
 * GNU Lesser General Public License (LGPL) version 2.1.
 * See the COPYING file in the parent directory for full text.
 */
package org.gearman.common;

import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.gearman.GearmanException;
import org.gearman.client.GearmanJobEventListener;
import org.gearman.client.GearmanClientImpl;

public class GearmanServerRequest implements GearmanJobEventListener{

    public static enum RequestState {
        NEW, AWAITINGRESPONSE, PROCESSINGRESPONSES, FINISHED,
        EXCEPTION
    }
    
    static final String DESCRIPTION_PREFIX = "GearmanServerRequest";
    private final String DESCRIPTION;
    private RequestState state = null;
    private GearmanPacket requestPacket = null;
    private UUID uuid = null;
    private GearmanServerResponseHandler handler = null;
    private static final Logger LOG = GearmanClientImpl.getLogger();

    public GearmanServerRequest(GearmanPacket request) {
        this(null,request);
    }



    public GearmanServerRequest(GearmanServerResponseHandler handler, GearmanPacket request) {
        if (request == null) {
            throw new IllegalArgumentException("The request can not be null");
        }

        if (!request.getMagic().equals(PacketMagic.REQ)) {
            throw new IllegalArgumentException("Invalid request. The packet " +
                    "is not a request packet");
        }

        state = RequestState.NEW;
        uuid = UUID.randomUUID();
        requestPacket = request; //TODO make a copy of the packet or make sure the packet is hardened against tampering
        this.handler = handler;
        DESCRIPTION = new String(DESCRIPTION_PREFIX + ":" + handler + ":" + uuid);

    }

    public void handleGearmanJobEvent(GearmanPacket p) throws IllegalArgumentException {
        if (p == null) {
            throw new IllegalArgumentException("You can not add a null" +
                    " response packet");
        }
        switch (state) {

                // request needs to be submitted to server
                case NEW:
                    if (p.getMagic().equals(PacketMagic.REQ)) {
                        if (requestPacket.requiresResponse()) {
                            changeState(RequestState.AWAITINGRESPONSE);
                        } else {
                            changeState(RequestState.FINISHED);
                        }
                    } else {
                        //TODO
                    }
                    break;

                case AWAITINGRESPONSE:
                    if (p.getMagic().equals(PacketMagic.RES)) {
                        changeState(RequestState.PROCESSINGRESPONSES);
                    } else {
                        //TODO (for now we break to avoid falling through)
                        break;
                    }

                case PROCESSINGRESPONSES:
                    if (!p.getMagic().equals(PacketMagic.RES)) {
                        //TODO
                        break;
                    }
                    if (handler == null) {
                        LOG.log(Level.WARNING,"ServerRequest requires " +
                                "response, but not response handler was " +
                                "provided for the request. Request = " + this);
                        changeState(RequestState.EXCEPTION);
                    }
                    try {
                        handler.handleEvent(p);
                        if (!handler.isDone()) {
                            changeState(RequestState.PROCESSINGRESPONSES);
                            return;
                        }
                    } catch (GearmanException ge) {
                        changeState(RequestState.EXCEPTION);
                        break;
                    }
                    changeState(RequestState.FINISHED);

                case FINISHED:
                    return;

                case EXCEPTION:
                    throw new GearmanException("Encountered Fatal Exception" +
                            " while driving the following ServerRequest: " +
                            requestPacket);

                default:
                    throw new GearmanException("Unknown Request state " +
                            state + " for request " + requestPacket);
            }
    }

    public GearmanPacket getRequestPacket() {
        return requestPacket;
    }

    public GearmanServerResponseHandler getHandler() {
        return handler;
    }

    public String toString() {
        return DESCRIPTION;
    }

    public RequestState getState() {
        return state;
    }

    private void changeState(RequestState newState) {
        if (!newState.equals(state)) {
            LOG.log(Level.FINE, "Request " + this + " is changing state" +
                    " from " + this.state + " to " + newState);
            this.state = newState;
        }
    }
}
