/*
 * Copyright (C) 2009 by Eric Lambert <Eric.Lambert@sun.com>
 * Use and distribution licensed under the
 * GNU Lesser General Public License (LGPL) version 2.1.
 * See the COPYING file in the parent directory for full text.
 */
package org.gearman.worker;


import org.gearman.common.PacketMagic;
import org.gearman.common.PacketType;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.gearman.Constants;
import org.gearman.GearmanException;
import org.gearman.worker.GearmanFunction;
import org.gearman.worker.GearmanFunctionFactory;
import org.gearman.common.GearmanPacket;
import org.gearman.common.GearmanJobServerConnection;
import org.gearman.common.GearmanServerRequest;
import org.gearman.common.GearmanServerResponseHandler;
import org.gearman.worker.GearmanWorker;
import org.gearman.common.Packet;
import org.gearman.util.ByteUtils;
import org.gearman.util.IORuntimeException;
public class GearmanWorkerImpl implements GearmanWorker {

    public enum State {

        IDLE, RUNNING, SHUTTINGDOWN
    }
    private static final String DESCRIPION_PREFIX = "GearmanWorker";
    private final String DESCRIPTION;
    private ArrayList<GearmanWorkerJobServerSession> sessionList = null;
    private ArrayList<GearmanFunction> functionList = null;
    private Selector ioAvailable = null;
    private static final Logger LOG = Logger.getLogger(
            Constants.GEARMAN_CLIENT_LOGGER_NAME);
    private String id;
    private HashMap<String, FunctionDefinition> functionMap;
    private State state;
    private ExecutorService executorService;

    class GrabJobEventHandler implements GearmanServerResponseHandler {

        private GearmanWorkerJobServerSession session;
        private boolean isDone = false;

        GrabJobEventHandler(GearmanWorkerJobServerSession session) {
            super();
            this.session = session;
        }

        public void handleEvent(GearmanPacket event) throws GearmanException {
            GearmanServerRequest req = null;
            switch (event.getPacketType()) {
                case JOB_ASSIGN:
                case JOB_ASSIGN_UNIQ:
                    if (!session.isAwake()) {
                        LOG.log(Level.WARNING, "Session " + session + " is " +
                                "asleep but has received a " + event.getPacketType() +
                                "response from job server.");
                        req = new GearmanServerRequest(
                                new GrabJobEventHandler(session),
                                new Packet(PacketMagic.REQ,
                                PacketType.WORK_FAIL,
                                event.getDataComponentValue(
                                GearmanPacket.DataComponentName.JOB_HANDLE)));
                    } else {
                        LOG.log(Level.INFO, "Session " + session +
                                " received " + event.getPacketType() +
                                " for job " + event);
                        session.addNewJob(event);
                        req = new GearmanServerRequest(
                                new GrabJobEventHandler(session),
                                new Packet(PacketMagic.REQ,
                                PacketType.GRAB_JOB, new byte[0]));
                        isDone = true;
                    }
                    break;

                case NO_JOB:
                    if (session.isAwake()) {
                        LOG.log(Level.FINE, "Session " + session +
                                " received NO_JOB response from the Job " +
                                "Server. Will send session to sleep.");
                        req = new GearmanServerRequest(
                                new GrabJobEventHandler(session),
                                new Packet(PacketMagic.REQ,
                                PacketType.PRE_SLEEP, new byte[0]));
                        session.setToSleep();
                        isDone = true;
                    } else {
                        LOG.log(Level.WARNING, "Session " + session +
                                " received an unsolicited NO_JOB response " +
                                "from the Job Server.");
                    }
                    break;

                case NOOP:
                    if (!session.isAwake()) {
                        LOG.log(Level.FINE, "Session " + session +
                                " received NOOP response from the Job Server." +
                                " Will wake session.");
                        req = new GearmanServerRequest(
                                new GrabJobEventHandler(session),
                                new Packet(PacketMagic.REQ,
                                PacketType.GRAB_JOB, new byte[0]));
                        session.wakeSession();
                        isDone = true;
                    } else {
                        LOG.log(Level.WARNING, "Session " + session +
                                " received an unsolicited NOOP response from " +
                                "the Job Server.");
                    }
                    break;

                case ERROR:
                    LOG.log(Level.SEVERE, "Session " + session +
                            " received an error packet from Job Server. " +
                            "Shutting down session.");
                    session.closeSession();
                    break;

                default:
                    LOG.log(Level.WARNING, "Session " + session +
                            " received an unknown response from the " +
                            "Job Server: " + event + ".");
            }
            if (req != null) {
                session.submitRequest(req);
            }
        }

        public boolean isDone() {
            return isDone;
        }
    }

    class GearmanFunctionNameFactory implements GearmanFunctionFactory {

        private String className = null;

        GearmanFunctionNameFactory(String className) {
            this.className = className;
        }

        public String getFunctionName() {
            return className;
        }

        public GearmanFunction getFunction() {
            GearmanFunction f = null;
            try {
                Class c = Class.forName(className);
                Object o = c.newInstance();
                if (o instanceof GearmanFunction) {
                    f = (GearmanFunction) o;
                }
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Unable to create instance of " +
                        "Function: " + className, e);
            }
            return f;
        }
    }

    class FunctionDefinition {

        private long timeout = 0;
        private GearmanFunctionFactory factory = null;

        FunctionDefinition(long timeout, GearmanFunctionFactory factory) {
            this.timeout = timeout;
            this.factory = factory;
        }

        long getTimeout() {
            return timeout;
        }

        GearmanFunctionFactory getFactory() {
            return factory;
        }
    }

    public GearmanWorkerImpl() {
        this(Executors.newSingleThreadExecutor());
    }

    public GearmanWorkerImpl(ExecutorService executorService) {
        DESCRIPTION = DESCRIPION_PREFIX + ":" + Thread.currentThread().getId();
        sessionList = new ArrayList<GearmanWorkerJobServerSession>();
        functionList = new ArrayList<GearmanFunction>();
        id = DESCRIPTION;
        functionMap = new HashMap<String, FunctionDefinition>();
        state = State.IDLE;
        this.executorService = executorService;
    }

    public String toString() {
        return id;
    }

    public void work() {
        if (!state.equals(State.IDLE)) {
            throw new IllegalStateException("Can not call work while worker " +
                    "is running or shutting down");
        }

        state = State.RUNNING;
        GearmanPacket p = new Packet(PacketMagic.REQ, PacketType.GRAB_JOB,
                new byte[0]);
        for (GearmanWorkerJobServerSession sess : sessionList) {
            GearmanServerRequest gsr = new GearmanServerRequest(
                    new GrabJobEventHandler(sess), p);
            sess.submitRequest(gsr);
        }
        while (isRunning()) {

            for (GearmanWorkerJobServerSession sess : sessionList) {
                try {
                    sess.driveSessionIO();
                } catch (IOException ioe) {
                    LOG.log(Level.WARNING, "Received IOException while driving" +
                            " IO on session " + sess, ioe);
                    continue;
                    //TODO, drop session?
                }
                Collection<GearmanPacket> assignedJobs = sess.getNewJobs();
                for (GearmanPacket curJob : assignedJobs) {
                    byte[] handle, data, functionNameBytes;
                    String functionName;
                    handle = curJob.getDataComponentValue(
                            GearmanPacket.DataComponentName.JOB_HANDLE);
                    functionNameBytes = curJob.getDataComponentValue(
                            GearmanPacket.DataComponentName.FUNCTION_NAME);
                    data = curJob.getDataComponentValue(
                            GearmanPacket.DataComponentName.DATA);
                    functionName = ByteUtils.fromUTF8Bytes(functionNameBytes);
                    FunctionDefinition def = functionMap.get(functionName);
                    if (def == null) {
                        GearmanServerRequest gsr = new GearmanServerRequest(
                                new Packet(PacketMagic.REQ,
                                PacketType.WORK_FAIL, handle));
                        sess.sendRequestToServer(gsr);
                    }
                    GearmanFunction function = def.getFactory().getFunction();
                    function.setData(data);
                    function.setJobHandle(handle);
                    function.registerEventListener(sess);
                    functionList.add(function);
                }
            }

            for (GearmanFunction fun : functionList) {
                Future<GearmanPacket> gp = executorService.submit(fun);
                FunctionDefinition def = functionMap.get(fun.getName());
                try {
                    if (def.getTimeout() > 0) {
                        gp.get(functionMap.get(fun.getName()).getTimeout(),
                                TimeUnit.MILLISECONDS);
                    } else {
                        gp.get();
                    }
                } catch (Exception e) {
                    fun.fireEvent(new Packet(PacketMagic.REQ,
                            PacketType.WORK_EXCEPTION,
                            Packet.generatePacketData(fun.getJobHandle(),
                            e.getMessage().getBytes())));
                    fun.fireEvent(new Packet(PacketMagic.REQ,
                            PacketType.WORK_FAIL, fun.getJobHandle()));;
                }
            }
            functionList.clear();

            boolean shouldSleep = true;
            for (GearmanWorkerJobServerSession sess : sessionList) {
                if (sess.isAwake()) {
                    shouldSleep = false;
                }
            }
            if (shouldSleep) {
                try {
                    Thread.currentThread().sleep(250);
                } catch (InterruptedException ignore) {
                }
            }
        }

        shutDownWorker(true);

    }

    public void addServer(String hostname)
            throws IllegalArgumentException, IllegalStateException {
        addServer(hostname, Constants.GEARMAN_DEFAULT_TCP_PORT);
    }

    public void addServer(String hostname, int port)
            throws IllegalArgumentException, IllegalStateException {

        InetSocketAddress sa = new InetSocketAddress(hostname, port);

        //this is a sub-optimal way to look for dups, but addJobServer
        //ops should be infrequent enough that this should be a big penalty
        Iterator<GearmanWorkerJobServerSession> iter = sessionList.iterator();
        while (iter.hasNext()) {
            if (iter.next().getServer().equals(sa)) {
                return;
            }
        }

        GearmanWorkerJobServerSession session = new GearmanWorkerJobServerSession(sa);
        sessionList.add(session);
        if (ioAvailable == null) {
            try {
                ioAvailable = Selector.open();
            } catch (IOException ioe) {
                throw new IORuntimeException(ioe);
            }
        }
        try {
            session.initSession(ioAvailable, SelectionKey.OP_READ | SelectionKey.OP_WRITE);
        } catch (IOException ioe) {
            throw new IORuntimeException(ioe);
        }

        GearmanPacket p = new Packet(PacketMagic.REQ, PacketType.SET_CLIENT_ID,
                ByteUtils.toUTF8Bytes(id));
        session.sendRequestToServer(new GearmanServerRequest(p));

        for (FunctionDefinition def : functionMap.values()) {
            p = generateCanDoPacket(def);
            session.sendRequestToServer(new GearmanServerRequest(p));
        }

        if (isRunning()) {
            p = new Packet(PacketMagic.REQ, PacketType.GRAB_JOB, new byte[0]);
            GearmanServerRequest gsr = new GearmanServerRequest(
                    new GrabJobEventHandler(session), p);
            session.sendRequestToServer(gsr);
        }
        LOG.log(Level.FINE, "Added server " + sa + " to worker " + this);
    }

    public void addServer(GearmanJobServerConnection conn) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public boolean hasServer(String hostname) {
        return hasServer(hostname, Constants.GEARMAN_DEFAULT_TCP_PORT);
    }

    public boolean hasServer(String hostname, int port) {
        InetSocketAddress sa = new InetSocketAddress(hostname, port);
        boolean foundIt = false;
        for (GearmanWorkerJobServerSession sess : sessionList) {
            if (sess.getServer().equals(sa)) {
                foundIt = true;
            }
        }
        return foundIt;
    }

    public boolean hasServer(GearmanJobServerConnection conn) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public String echo(String text, GearmanJobServerConnection conn) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void registerFunction(String function, long timeout) {
        registerFunctionFactory(new GearmanFunctionNameFactory(function), timeout);
    }

    public void registerFunction(String function) {
        registerFunction(function, 0);
    }

    public void registerFunction(Class<? extends GearmanFunction> function) {
        registerFunction(function, 0);
    }

    public void registerFunction(Class<? extends GearmanFunction> function, long timeout) {
        registerFunctionFactory(new GearmanFunctionNameFactory(
                function.getCanonicalName()), timeout);
    }

    public void registerFunctionFactory(GearmanFunctionFactory factory) {
        registerFunctionFactory(factory, 0);
    }

    public void registerFunctionFactory(GearmanFunctionFactory factory, long timeout) {
        if (functionMap.containsKey(factory.getFunctionName())) {
            return;
        }
        FunctionDefinition def = new FunctionDefinition(timeout, factory);
        functionMap.put(factory.getFunctionName(), def);
        sendToAll(generateCanDoPacket(def));
    }

    public Set<String> getRegisteredFunctions() {
        HashSet<String> functions = new HashSet<String>();
        for (FunctionDefinition def : functionMap.values()) {
            functions.add(def.factory.getFunctionName());
        }
        return functions;
    }

    public void setWorkerID(String id) throws IllegalArgumentException {
        if (id == null) {
            throw new IllegalArgumentException("Worker ID may not be null");
        }
        this.id = id;
        sendToAll(new Packet(PacketMagic.REQ, PacketType.SET_CLIENT_ID,
                ByteUtils.toUTF8Bytes(id)));
    }

    public void setWorkerID(String id, GearmanJobServerConnection conn) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public String getWorkerID() {
        return id;
    }

    public String getWorkerID(GearmanJobServerConnection conn) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void unregisterFunction(String functionName) {
        functionMap.remove(functionName);
        sendToAll(new Packet(PacketMagic.REQ, PacketType.CANT_DO,
                ByteUtils.toUTF8Bytes(functionName)));
    }

    public void unregisterAll() {
        functionMap.clear();
        sendToAll(new Packet(PacketMagic.REQ, PacketType.RESET_ABILITIES,
                new byte[0]));
    }

    public void stop() {
        state = State.SHUTTINGDOWN;
    }

    public List<Exception> shutdown() {
        return shutDownWorker(false);
    }

    public boolean isRunning() {
        return state.equals(State.RUNNING);
    }

    private GearmanPacket generateCanDoPacket(FunctionDefinition def) {
        PacketType pt = PacketType.CAN_DO;
        byte[] data = null;
        byte[] name = ByteUtils.toUTF8Bytes(def.getFactory().getFunctionName());
        long timeout = def.getTimeout();

        if (timeout > 0) {
            pt = PacketType.CAN_DO_TIMEOUT;
            byte[] to = ByteUtils.toUTF8Bytes(String.valueOf(timeout));
            data = new byte[name.length + to.length + 1];
            System.arraycopy(name, 0, data, 0, name.length);
            data[name.length] = ByteUtils.NULL;
            System.arraycopy(to, 0, data, name.length + 1, to.length);
        } else {
            data = name;
        }
        return new Packet(PacketMagic.REQ, pt, data);
    }

    private void sendToAll(GearmanPacket p) {
        sendToAll(null, p);
    }

    private void sendToAll(GearmanServerResponseHandler handler, GearmanPacket p) {
        GearmanServerRequest gsr = null;
        if (handler == null) {
            gsr = new GearmanServerRequest(p);
        } else {
            gsr = new GearmanServerRequest(handler, p);
        }
        for (GearmanWorkerJobServerSession sess : sessionList) {
            sess.sendRequestToServer(gsr);
        }

    }

    /*
     * For the time being this will always return an empty list of
     * exceptions because closeSession does not throw an exception
     */
    private List<Exception> shutDownWorker(boolean completeTasks) {

        ArrayList<Exception> exceptions = new ArrayList<Exception>();
        // by setting the session to sleep, we will reject any Job assigns but
        // still leave the session usable by anyjobs in flight.
        for (GearmanWorkerJobServerSession sess : sessionList) {
            sess.setToSleep();
        }

        // This gives any jobs in flight a chance to complete
        if (completeTasks) {
            executorService.shutdown();
        } else {
            executorService.shutdownNow();
        }

        for (GearmanWorkerJobServerSession sess : sessionList) {
            sess.closeSession();
        }
        state = State.IDLE;

        return exceptions;
    }
}
